part of ormicida.codec.bson;

class BsonDecoder implements CodecDecoder {
  CodecListener       _listener;
  Buffer              _input;
  final StringBuffer  _buffer = new StringBuffer();
  
  BsonDecoder();
  
  void clear() {
    _listener = null;
    _input = null;
    _buffer.clear();
  }
  
  void decodeTo(Buffer input, CodecListener listener) {
    clear();
    _input = input;
    _listener = listener;
    input.setLittleEndian();
    
    _readObject();
  }
  
  void _readObject() {
    final end = _input.position + _input.readUint32();
    _listener.onMapStart();
    
    while (true) {
      final type = _input.readByte();
      if (type == 0) break;
      
      final buffer = _buffer;
      buffer.clear();
      while (true) {
        final code = _input.readByte();
        if (code == 0) break;
        buffer.writeCharCode(code);
      }
      _listener.onPropertyStart(buffer.toString());
      buffer.clear();
      _readValue(type);
      _listener.onPropertyEnd();
    }
    
    _listener.onMapEnd();
  }
  
  void _readValue(int type) {
    final input = _input;
    
    switch(type) {
      case _STRING:
        final lengthInBytes = input.readUint32();
        final buffer = _buffer;
        buffer.clear();
        while (true) {
          final code = input.readCharCode();
          if (code == 0) break;
          buffer.writeCharCode(code);
        }
        _listener.onValue(buffer.toString());
        buffer.clear();
        break;
        
      case _OBJECT_ID:
        _listener.onValue(new SchemaId(input.readBytes(12)));
        break;
        
      case _OBJECT:
        _readObject();
        break;
        
      case _ARRAY:
        final length = input.readUint32();
        _listener.onListStart();
        
        while (true) {
          final type = input.readByte();
          if (type == 0) break;
          
          // Skip index
          while (true) {
            final code = input.readByte();
            if (code == 0) break;
          }
          
          _readValue(type);
          _listener.onListElement();
        }
        
        _listener.onListEnd();
        break;
        
      case _INT32:
        _listener.onValue(input.readInt32());
        break;
        
      case _INT64:
        _listener.onValue(input.readInt64());
        break;
        
      case _NULL:
        _listener.onValue(null);
        break;
        
      case _BOOLEAN:
        _listener.onValue(input.readByte() > 0);
        break;
        
      case _FLOAT:
        _listener.onValue(input.readFloat64());
        break;
        
      case _DATE_TIME:
        _listener.onValue(
            new DateTime.fromMillisecondsSinceEpoch(
                input.readInt64()));
        break;
        
      case _BINARY:
        final length = input.readUint32();
        final binaryType = input.readByte();
        _listener.onValue(input.readBytes(length));
        break;
        
      case _REGEXP:
        final buffer = _buffer;
        buffer.clear();
        while (true) {
          final code = input.readByte();
          if (code == 0) break;
          buffer.writeCharCode(code);
        }
        final pattern = buffer.toString();
        buffer.clear();
        
        bool isCaseSensitive = true;
        bool isMultiLine = false;
        while (true) {
          final code = input.readByte();
          if (code == 0) break;
          switch (code) {
            case _CHAR_i:
              isCaseSensitive = false;
              break;
            case _CHAR_m:
              isMultiLine = true;
              break;
          }
        }
        
        _listener.onValue(new RegExp(pattern,
            caseSensitive: isCaseSensitive,
            multiLine: isMultiLine));
        break;
        
      case _TIMESTAMP:
        _listener.onValue(input.readUint64());
        break;
        
      default:
        throw new UnimplementedError('Unimplemented BSON type: 0x${type.toRadixString(16)}');
    }
  }
}
